"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.defaultConfig = exports.Preszr = void 0;
const utils_1 = require("./utils");
const data_1 = require("./data");
const encodings_1 = require("./encodings");
const errors_1 = require("./errors");
const unsupported_1 = require("./unsupported");
const store_1 = require("./encode/store");
const encode_context_1 = require("./encode/encode-context");
const fixed_1 = require("./encodings/fixed");
const decode_context_1 = require("./encode/decode-context");
/**
 * The class used to encode and decode things in the preszr format.
 */
class Preszr {
    constructor(config) {
        this.config = exports.defaultConfig;
        this._store = new store_1.EncodingStore();
        this.config = (0, utils_1.defaultsDeep)((0, utils_1.cloneDeep)(config), exports.defaultConfig);
        const unsupportedEncoding = (0, encodings_1.getUnsupportedEncoding)(...unsupported_1.unsupportedTypes, ...this.config.unsupported);
        this._store.add(...encodings_1.builtinEncodings, unsupportedEncoding, ...this.config.encodes);
    }
    _checkInputHeader(input) {
        let reason = "";
        let versionInfo = "";
        if (!Array.isArray(input)) {
            reason = "input is not array";
        }
        else {
            const header = input === null || input === void 0 ? void 0 : input[0];
            if (!header) {
                reason = "no header element";
            }
            else if (!Array.isArray(header)) {
                reason = "header element is not array";
            }
            else {
                versionInfo = header[0];
                if (versionInfo == null) {
                    reason = "no version info";
                }
                else if (typeof versionInfo !== "string") {
                    reason = "version is not string";
                }
                else if (+versionInfo !== parseInt(versionInfo)) {
                    reason = "version is not numeric";
                }
                else if (versionInfo !== utils_1.version) {
                    throw new errors_1.PreszrError("Decoding", `Input was encoded using version ${versionInfo}, but preszr is version ${utils_1.version}. Set skipValidateVersion to allow this.`);
                }
                else if (!Array.isArray(header[1])) {
                    reason = "no encoding keys or encoding keys not an array";
                }
                else if (typeof header[2] !== "object" || !header[1]) {
                    reason =
                        "no encoding data or encoding data is not an object";
                }
                else if (typeof header[3] !== "object" || !header[2]) {
                    reason =
                        "no custom metadata or custom metadata is not an object";
                }
                else if (input.length === 1) {
                    reason = "input must have at least 2 elements";
                }
            }
        }
        if (reason) {
            throw new errors_1.PreszrError("Decoding", `Input is not preszr-encoded: ${reason}`);
        }
    }
    decode(input) {
        const tryScalar = (0, data_1.tryDecodeScalar)(input);
        if (tryScalar !== data_1.noResultPlaceholder)
            return tryScalar;
        input = input;
        // We check the header is in Preszr format.
        this._checkInputHeader(input);
        // Deconstruct the header to its parts
        const [[, encodingKeys, encodingSpec, metadata]] = input;
        const targetArray = Array(input.length - 1);
        const needToInit = new Map();
        // For optimization purposes, create one instance of `InitContext`.
        const ctx = new decode_context_1.DecodeContext(targetArray);
        const encodingByIndex = [];
        // Check all encoding keys are present.
        for (const encodingKey of encodingKeys) {
            encodingByIndex.push(this._store.mustGetByKey(encodingKey));
        }
        // Start decoding the payload.
        for (let i = 1; i < input.length; i++) {
            const encodingIndex = encodingSpec[i];
            const cur = input[i];
            let encoding;
            if (encodingIndex == null) {
                if (typeof cur === "string") {
                    targetArray[i] = cur;
                    continue;
                }
                if (Array.isArray(cur)) {
                    encoding = encodings_1.arrayEncoding;
                }
                else {
                    encoding = encodings_1.objectEncoding;
                }
            }
            else if (encodingIndex < fixed_1.Fixed.End) {
                encoding = this._store.mustGetByIndex(encodingIndex);
            }
            else {
                encoding = encodingByIndex[encodingIndex - fixed_1.Fixed.End];
            }
            if (encodingIndex === fixed_1.Fixed.UnrecognizedSymbol) {
                targetArray[i] = (0, utils_1.getUnrecognizedSymbol)(metadata[i]);
                continue;
            }
            if ("symbol" in encoding) {
                targetArray[i] = encoding.symbol;
                continue;
            }
            const protoEnc = encoding;
            ctx.metadata = metadata[i];
            targetArray[i] = protoEnc.decoder.create(cur, ctx);
            if (protoEnc.decoder.init) {
                needToInit.set(i, protoEnc);
            }
        }
        ctx.next();
        for (const key of needToInit.keys()) {
            const encoding = needToInit.get(key);
            ctx.metadata = metadata[key];
            encoding.decoder.init(targetArray[key], input[key], ctx);
        }
        return targetArray[1];
    }
    encode(root) {
        const tryScalar = (0, data_1.tryEncodeScalar)(root);
        if (tryScalar !== data_1.noResultPlaceholder) {
            return tryScalar;
        }
        const ctx = new encode_context_1.EncodeCtx(this._store);
        ctx.encode(root);
        const result = ctx.finish();
        return result;
    }
}
exports.Preszr = Preszr;
exports.defaultConfig = {
    encodes: [],
    unsupported: []
};
//# sourceMappingURL=core.js.map