"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.EncodingStore = void 0;
const interface_1 = require("../interface");
const errors_1 = require("../errors");
const utils_1 = require("../encodings/utils");
const utils_2 = require("../utils");
const fixed_1 = require("../encodings/fixed");
const encodings_1 = require("../encodings");
class EncodingStore {
    constructor() {
        // We use multiple caches to speed up encoding and decoding.
        // It's a space/time trade-off, but it'll barely take any space.
        // This maintains a full list of all versions of all prototype encodings.
        // it's used for various checks and to built the caches.
        this._nameToProtoEncodings = new Map();
        // This maintains all symbol encodings and is also used when encoding.
        this._symbolToEncoding = new Map();
        // Matches a prototype to an encoding. Used for some checks.
        this._protoToEncoding = new Map();
        // A cache that quickly matches an encoding key, e.g. Custom/whatever.v5 -> encoding.
        // used when decoding.
        // Both symbol and proto encodings use this cache, but their keys are different.
        // Proto encodings have keys ending with .v${version}, and symbol encodings have keys
        // ending with .S
        this._keyToEncoding = new Map();
        // A "map" of index to built-in encoding. Has empty indexes.
        this._indexToEncoding = Array(fixed_1.Fixed.End);
    }
    all() {
        return [...this.getProtoEncodings(), ...this.getSymbolEncodings()];
    }
    _registerProtos(encoding) {
        for (const proto of encoding.protos) {
            const existingEncoding = this._protoToEncoding.get(proto);
            // It's illegal to have two different encodings for the same prototype, since it becomes ambiguous.
            if (existingEncoding && existingEncoding.name !== encoding.name) {
                throw new errors_1.PreszrError("Configuration", `Encoding ${(0, utils_1.getEncodingKey)(encoding)} references prototype ${(0, utils_2.getClassName)(proto)}, but encoding ${(0, utils_1.getEncodingKey)(existingEncoding)} also refers that prototype.`);
            }
            // However, it's okay to have the same encoding with different versions reference different prototypes.
            if (!existingEncoding ||
                existingEncoding.version < encoding.version) {
                this._protoToEncoding.set(proto, encoding);
            }
        }
    }
    add(...specs) {
        for (const spec of specs) {
            const encoding = (0, utils_1.makeFullEncoding)(spec);
            // After adding encodings, the cache must be rebuilt.
            if ("symbol" in encoding) {
                this._addSymbolEncoding(encoding);
            }
            else {
                this._addProtoEncoding(encoding);
            }
            const fixed = encoding[interface_1.fixedIndexProp];
            if (fixed != null) {
                const existing = this._indexToEncoding[fixed];
                if (existing && existing.name !== encoding.name) {
                    const enc1 = (0, utils_1.getEncodingKey)(encoding);
                    const enc2 = (0, utils_1.getEncodingKey)(existing);
                    throw new errors_1.PreszrError("Configuration", `Encodings '${enc1}' and '${enc2}' have identical fixed index (${fixed}).`);
                }
                this._indexToEncoding[fixed] = encoding;
            }
        }
    }
    _addProtoEncoding(encoding) {
        let versioned = this._nameToProtoEncodings.get(encoding.name);
        if (!versioned) {
            // Initialize the versioned encoding list if it doesn't exist
            versioned = new Map();
            this._nameToProtoEncodings.set(encoding.name, versioned);
        }
        if (versioned.has(encoding.version)) {
            throw new errors_1.PreszrError("Configuration", `Encoding ${encoding.name} with version ${encoding.version} already exists in this instance.`);
        }
        this._registerProtos(encoding);
        // The latest encoding version is kept under -1 for easy access.
        const maxVersionEncoding = versioned.get(-1);
        if (!maxVersionEncoding ||
            maxVersionEncoding.version < encoding.version) {
            versioned.set(-1, encoding);
            // If the max version was set, that means the proto cache is obsolete
            this._cacheProtoToEncoding = undefined;
        }
        versioned.set(encoding.version, encoding);
        this._keyToEncoding.set((0, utils_1.getEncodingKey)(encoding), encoding);
    }
    *getProtoEncodings() {
        for (const [name, versions] of this._nameToProtoEncodings) {
            // get latest version encoding
            const maxVersionEncoding = versions.get(-1);
            if (!maxVersionEncoding) {
                throw new errors_1.PreszrError("Bug", `Expected there to be a maximum version encoding for ${name}.`);
            }
            yield maxVersionEncoding;
        }
    }
    getSymbolEncodings() {
        return this._symbolToEncoding.values();
    }
    _addSymbolEncoding(encoding) {
        const existingBySymbol = this._symbolToEncoding.get(encoding.symbol);
        const existingByKey = this._keyToEncoding.get((0, utils_1.getEncodingKey)(encoding));
        if (existingByKey) {
            throw new errors_1.PreszrError("Configuration", `Encoding with the name ${encoding.name} already exists in this instance.`);
        }
        if (existingBySymbol) {
            throw new errors_1.PreszrError("Configuration", `${encoding.name} references symbol ${(0, utils_2.getSymbolName)(encoding.symbol)}, but encoding ${existingBySymbol.name} also references that symbol.`);
        }
        this._symbolToEncoding.set(encoding.symbol, encoding);
        this._keyToEncoding.set((0, utils_1.getEncodingKey)(encoding), encoding);
    }
    mayGetBySymbol(s) {
        return this._symbolToEncoding.get(s);
    }
    mustGetByIndex(ix) {
        const encoding = this._indexToEncoding[ix];
        if (!encoding) {
            throw new errors_1.PreszrError("Decoding", `Encoding index ${ix} doesn't refer to any known encoding.`);
        }
        return encoding;
    }
    mustGetByKey(key) {
        var _a;
        const encoding = this._keyToEncoding.get(key);
        if (encoding) {
            return encoding;
        }
        // If it failed, this is going to be an error
        // which means there are no performance considerations. Let's gather as much
        // info as possible.
        const info = (0, utils_1.mustParseEncodingKey)(key);
        if (info.type === "symbol") {
            throw new errors_1.PreszrError("Decoding", `No symbol encoding for name ${info.name}.`);
        }
        else {
            const namedProtoEncoding = (_a = this._nameToProtoEncodings
                .get(info.name)) === null || _a === void 0 ? void 0 : _a.get(-1);
            if (!namedProtoEncoding) {
                throw new errors_1.PreszrError("Decoding", `No prototype encoding named ${info.name}, for any version.`);
            }
            throw new errors_1.PreszrError("Decoding", `Prototype encoding ${info.name} doesn't exist for version ${info.version}. There is an encoding for ${namedProtoEncoding.version}`);
        }
    }
    _makeWorkingProtoCache() {
        // We add an entry to the working encoding cache for every registered
        // encoding.
        const cache = new WeakMap();
        for (const encoding of this.getProtoEncodings()) {
            if (!encoding) {
                throw new errors_1.PreszrError("Bug", `Expected there to be a maximum version encoding for ${encoding.name}.`);
            }
            for (const proto of encoding.protos) {
                cache.set(proto, encoding);
            }
        }
        return cache;
    }
    mustGetByProto(obj) {
        var _a;
        if (!this._cacheProtoToEncoding) {
            this._cacheProtoToEncoding = this._makeWorkingProtoCache();
        }
        const fromCache = this._cacheProtoToEncoding.get(obj);
        if (fromCache) {
            return fromCache;
        }
        let foundEncoding;
        const chain = [];
        for (let proto = obj;; proto = (_a = Object.getPrototypeOf(proto)) !== null && _a !== void 0 ? _a : encodings_1.nullPlaceholder) {
            const cached = this._cacheProtoToEncoding.get(proto);
            if (cached !== undefined) {
                foundEncoding = cached;
                break;
            }
            if (proto === encodings_1.nullPlaceholder) {
                throw new errors_1.PreszrError("Bug", "mustGetByProto failed to find anything.");
            }
            chain.push(proto);
        }
        for (const proto of chain) {
            this._cacheProtoToEncoding.set(proto, foundEncoding);
        }
        return foundEncoding;
    }
}
exports.EncodingStore = EncodingStore;
//# sourceMappingURL=store.js.map