"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.mustParseEncodingKey = exports.getEncodingKey = exports.makeFullEncoding = exports.makeProtoEncoding = exports.makeProtoEncodingByCtor = exports.makeSymbolEncoding = void 0;
const errors_1 = require("../errors");
const objects_1 = require("./objects");
const utils_1 = require("../utils");
const interface_1 = require("../interface");
const index_1 = require("./index");
const MAX_VERSION = 999;
const MIN_VERSION = 0;
function makeSymbolEncoding(x) {
    if (typeof x !== "symbol") {
        return x;
    }
    const name = (0, utils_1.getSymbolName)(x);
    if (!name) {
        throw new errors_1.PreszrError("Configuration", `Symbol has no name. You must specify a 'key'.`);
    }
    return {
        name: name,
        symbol: x
    };
}
exports.makeSymbolEncoding = makeSymbolEncoding;
function makeProtoEncodingByCtor(ctor) {
    if (!ctor.prototype) {
        throw new errors_1.PreszrError("Configuration", "Failed to detect prototype from constructor.");
    }
    return makeProtoEncoding({
        proto: ctor.prototype,
        version: 0
    });
}
exports.makeProtoEncodingByCtor = makeProtoEncodingByCtor;
function makeProtoEncoding(specifier) {
    var _a, _b, _c, _d;
    // protype CAN be `null`.
    if (specifier.proto === undefined) {
        throw new errors_1.PreszrError("Configuration", "Encoding must specify a prototype.");
    }
    if (typeof specifier.proto === "function") {
        throw new errors_1.PreszrError("Configuration", "Prototype can't be a function. Did you mean to supply a constructor instead?");
    }
    const v = specifier.version;
    if (v != null &&
        (typeof v !== "number" || !Number.isSafeInteger(v) || v < 1)) {
        throw new errors_1.PreszrError("Configuration", `Version for encoding ${specifier.name} must be an safe, positive integer, but was: ${v}.`);
    }
    const builtInEncoding = (0, index_1.getBuiltInEncoding)(specifier.proto);
    const encoding = {};
    if (builtInEncoding) {
        if (specifier.name != null) {
            throw new errors_1.PreszrError("Configuration", "When overriding a default encoding, 'name' must not exist.");
        }
        if (v == null || v <= 0) {
            throw new errors_1.PreszrError("Configuration", "When overriding a default encoding, 'version' must exist and be greater than 0.");
        }
        specifier.name = builtInEncoding.name;
        encoding[interface_1.fixedIndexProp] = builtInEncoding[interface_1.fixedIndexProp];
    }
    const proto = (_a = specifier.proto) !== null && _a !== void 0 ? _a : objects_1.nullPlaceholder;
    const className = (0, utils_1.getClassName)(proto);
    if (!className && !specifier.name) {
        throw new errors_1.PreszrError("Configuration", `No key has been provided, and the prototype has no name.`);
    }
    encoding.protos = [proto];
    encoding.name = (_b = specifier.name) !== null && _b !== void 0 ? _b : (0, utils_1.getImplicitClassEncodingName)(className);
    encoding.encode = (_c = specifier.encode) !== null && _c !== void 0 ? _c : (0, objects_1.getPrototypeEncoder)(proto);
    encoding.decoder = (_d = specifier.decoder) !== null && _d !== void 0 ? _d : (0, objects_1.getPrototypeDecoder)(proto);
    encoding.version = specifier.version == null ? 0 : specifier.version;
    return encoding;
}
exports.makeProtoEncoding = makeProtoEncoding;
function makeFullEncoding(specifier) {
    if (typeof specifier === "symbol" || "symbol" in specifier) {
        return makeSymbolEncoding(specifier);
    }
    if (typeof specifier === "function") {
        return makeProtoEncodingByCtor(specifier);
    }
    if ("proto" in specifier) {
        specifier = makeProtoEncoding(specifier);
    }
    if (!specifier.protos || specifier.protos.length === 0) {
        throw new errors_1.PreszrError("Configuration", "Encoding must specify prototypes.");
    }
    if (!("decoder" in specifier && "encode" in specifier)) {
        throw new errors_1.PreszrError("Configuration", "Multi-prototype specifier must have both decoder and encode.");
    }
    if (!specifier.name) {
        throw new errors_1.PreszrError("Configuration", "Multi-prototype specifier must provide a key.");
    }
    const v = specifier.version;
    if (typeof v !== "number" || !Number.isSafeInteger(v)) {
        throw new errors_1.PreszrError("Configuration", `Version for encoding ${specifier.name} must be an safe integer, but was: ${v}.`);
    }
    if (v > MAX_VERSION || v < MIN_VERSION) {
        throw new errors_1.PreszrError("Configuration", `Version number for ${specifier.name} must be between ${MIN_VERSION} and ${MAX_VERSION}, but was: ${v}`);
    }
    return specifier;
}
exports.makeFullEncoding = makeFullEncoding;
function getEncodingKey(enc) {
    if ("protos" in enc) {
        return `${enc.name}.v${enc.version}`;
    }
    return `${enc.name}.S`;
}
exports.getEncodingKey = getEncodingKey;
function mustParseEncodingKey(key) {
    const lastDot = key.lastIndexOf(".");
    if (lastDot === -1) {
        throw new errors_1.PreszrError("Decoding", `In encoding key ${key}, had no postfix.`);
    }
    const strPostfix = key.slice(lastDot + 1);
    const name = key.slice(0, lastDot);
    if (strPostfix === "S") {
        return {
            type: "symbol",
            name
        };
    }
    if (strPostfix !== "v") {
        throw new errors_1.PreszrError("Decoding", `Unknown postfix ${strPostfix}.`);
    }
    const strVersion = strPostfix.slice(1);
    if (!(0, utils_1.isNumericString)(strVersion)) {
        throw new errors_1.PreszrError("Decoding", `For encoding key ${key}, version wasn't numeric.`);
    }
    if (!name) {
        throw new errors_1.PreszrError("Decoding", `For encoding key ${key}, name was empty.`);
    }
    return {
        type: "prototype",
        name,
        version: +strVersion
    };
}
exports.mustParseEncodingKey = mustParseEncodingKey;
//# sourceMappingURL=utils.js.map