import { EncodedEntity, ScalarValue } from "./data";
export declare const fixedIndexProp: unique symbol;
/**
 * The context used by the encoding process.
 */
export interface EncodeContext {
    /**
     * Encodes the given input. For entities, it will recursively encode them, add
     * them to the final output as a side effect, and return a reference. For other
     * values, it will return them as-is or encode them, usually as a string.
     * @param value
     */
    encode(value: any): ScalarValue;
    /**
     * For internal use only.
     */
    _isImplicit: boolean;
    /**
     * Sets the metadata for this entity. The metadata can be any JSON-legal value,
     * including an object. It doesn't do anything, but can be accessed while decoding.
     */
    metadata: any;
}
/**
 * The context used by the create stage of the decoding process. Only
 * exposes the entity's metadata.
 */
export interface CreateContext {
    metadata: any;
}
/**
 * The context used by the init stage of the decoding process. Allows
 * resolving references to other entities.
 */
export interface InitContext {
    metadata: any;
    decode(value: ScalarValue): unknown;
}
/**
 * The decoding logic for a prototype encoding.
 */
export interface Decoder {
    create(encoded: EncodedEntity, ctx: CreateContext): unknown;
    init?(target: unknown, encoded: EncodedEntity, ctx: InitContext): void;
}
/**
 * Specifies a prototype encoding. Missing fields will be filled in automatically.
 */
export interface PrototypeEncodingSpecifier {
    name?: string;
    version?: number;
    proto: object | null;
    decoder?: Decoder;
    encode?(input: any, ctx: EncodeContext): EncodedEntity;
}
/**
 * This special encoding is used to encode missing symbols.
 */
export interface SpecialEncoding {
    name: string;
    [fixedIndexProp]?: number;
    metadata?: any;
}
/**
 * A full symbol encoding.
 */
export interface SymbolEncoding {
    name: string;
    symbol: symbol;
    [fixedIndexProp]?: number;
}
/**
 * A full prototype encoding.
 */
export interface PrototypeEncoding {
    name: string;
    version: number;
    [fixedIndexProp]?: number;
    protos: object[];
    decoder: Decoder;
    encode(input: any, ctx: EncodeContext): EncodedEntity;
}
export declare type UserEncoding = PrototypeEncoding | SymbolEncoding;
/**
 * A full preszr encoding of any type.
 */
export declare type Encoding = PrototypeEncoding | SymbolEncoding | SpecialEncoding;
/**
 * These encodings specifiers can't be confused for configuration objects or
 * prototypes.
 */
export declare type SimpleEncodingSpecifier = symbol | Function;
/**
 * An encoding specifier. Can be a symbol or constructor for a shorthand
 * symbol or prototype encoding. You can also give symbol or prototype encoding specifier
 * if you want to be more explicit.
 */
export declare type EncodingSpecifier = SimpleEncodingSpecifier | PrototypeEncodingSpecifier | PrototypeEncoding | SymbolEncoding;
/**
 * Configuration for an Preszr instance.
 */
export interface PreszrConfig {
    /**
     * An array of encoding specifiers. If you put your constructors and symbols here,
     * the Preszr will recognize them.
     */
    encodes: EncodingSpecifier[];
    /**
     * An array of constructors that will be marked as unsupported.
     * Objects with these constructors will not be encoded.
     */
    unsupported: Function[];
}
/**
 * Similar to Partial<T>, but recursively applied.
 */
export declare type DeepPartial<T> = {
    [K in keyof T]?: T[K] extends object ? DeepPartial<T[K]> : T[K];
};
//# sourceMappingURL=interface.d.ts.map